use crossterm::event::{KeyCode, KeyEvent, KeyEventKind};
use color_eyre::Result;
use ratatui::{
    layout::{Alignment, Constraint, Flex, Layout, Rect}, 
    text::Text, 
    widgets::{Block, Cell, Clear, Row, Table}, 
    Frame,
};
use tokio::sync::mpsc::UnboundedSender;
use std::sync::mpsc::Sender;

use super::{Component, PartialComponent, CurrentTab};
use crate::{
    action::Action, 
    config::Config, 
    palette::StylePalette, 
};

#[derive(Debug)]
pub struct RotatePopup {
    is_active: bool,
    action_tx: Option<UnboundedSender<Action>>,
    network_tx: Option<Sender<Action>>,
    cached_keys: String,
    secret_seed: [u8; 32],
    palette: StylePalette
}

impl Default for RotatePopup {
    fn default() -> Self {
        Self::new()
    }
}

impl RotatePopup {
    pub fn new() -> Self {
        Self {
            is_active: false,
            action_tx: None,
            network_tx: None,
            cached_keys: String::new(),
            secret_seed: [0u8; 32],
            palette: StylePalette::default(),
        }
    }

    fn rotate_keys(&mut self) {
        if !self.cached_keys.is_empty() && self.cached_keys.len() == 258 {
            if let Some(network_tx) = &self.network_tx {
                let _ = network_tx.send(Action::SetSessionKeys(
                        self.secret_seed, self.cached_keys.clone()));
            }
            if let Some(action_tx) = &self.action_tx {
                let _ = action_tx.send(Action::ClosePopup);
            }
        } else {
            if let Some(network_tx) = &self.network_tx {
                let _ = network_tx.send(Action::RotateSessionKeys);
            }
        }
    }

    fn parse_session_keys(&self) -> (String, String, String, String) {
        if !self.cached_keys.is_empty() && self.cached_keys.len() == 258 {
            let gran_key = format!("0x{}", &self.cached_keys[2..66]);
            let babe_key = format!("0x{}", &self.cached_keys[66..130]);
            let audi_key = format!("0x{}", &self.cached_keys[130..194]);
            let slow_key = format!("0x{}", &self.cached_keys[194..258]);

            (gran_key, babe_key, audi_key, slow_key)
        } else {
            (
                String::from("0x0000000000000000000000000000000000000000000000000000000000000000"),
                String::from("0x0000000000000000000000000000000000000000000000000000000000000000"),
                String::from("0x0000000000000000000000000000000000000000000000000000000000000000"),
                String::from("0x0000000000000000000000000000000000000000000000000000000000000000"),
            )
        }
    }
}

impl PartialComponent for RotatePopup {
    fn set_active(&mut self, current_tab: CurrentTab) {
        match current_tab {
            CurrentTab::RotatePopup => self.is_active = true,
            _ => self.is_active = false,
        };
    }
}

impl Component for RotatePopup {
    fn register_network_handler(&mut self, tx: Sender<Action>) -> Result<()> {
        self.network_tx = Some(tx);
        Ok(())
    }

    fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
        self.action_tx = Some(tx);
        Ok(())
    }

    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Wallet) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_popup_style(style.get("popup_style").copied());
            self.palette.with_popup_title_style(style.get("popup_title_style").copied());
        }
        Ok(())
    }

    fn update(&mut self, action: Action) -> Result<Option<Action>> {
        match action {
            Action::StoreRotatedKeys(cached_keys) => self.cached_keys = cached_keys,
            Action::SetStashSecret(secret_seed) => self.secret_seed = secret_seed,
            _ => {}
        };
        Ok(None)
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        if self.is_active && key.kind == KeyEventKind::Press {
            match key.code {
                KeyCode::Enter => self.rotate_keys(),
                KeyCode::Esc => self.is_active = false, 
                _ => {},
            };
        }
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        if self.is_active {
            let (border_style, border_type) = self.palette.create_popup_style();

            let (gran_key, babe_key, audi_key, slow_key) = self.parse_session_keys();
            let table = Table::new(
                vec![
                    Row::new(vec![
                        Cell::from(Text::from("gran".to_string()).alignment(Alignment::Left)),
                        Cell::from(Text::from(gran_key).alignment(Alignment::Right)), 
                    ]),
                    Row::new(vec![
                        Cell::from(Text::from("babe".to_string()).alignment(Alignment::Left)),
                        Cell::from(Text::from(babe_key).alignment(Alignment::Right)), 
                    ]),
                    Row::new(vec![
                        Cell::from(Text::from("audi".to_string()).alignment(Alignment::Left)),
                        Cell::from(Text::from(audi_key).alignment(Alignment::Right)), 
                    ]),
                    Row::new(vec![
                        Cell::from(Text::from("slow".to_string()).alignment(Alignment::Left)),
                        Cell::from(Text::from(slow_key).alignment(Alignment::Right)), 
                    ]),
                    ],
                [
                    Constraint::Length(4),
                    Constraint::Min(0),
                ],
            )
            .highlight_style(self.palette.create_highlight_style())
            .column_spacing(1)
            .block(Block::bordered()
                .border_style(border_style)
                .border_type(border_type)
                .title_alignment(Alignment::Right)
                .title_style(self.palette.create_title_style(false))
                .title("Rotate session keys (Enter to proceed / Esc to close)"));

            let v = Layout::vertical([Constraint::Max(6)]).flex(Flex::Center);
            let h = Layout::horizontal([Constraint::Max(73)]).flex(Flex::Center);
            let [area] = v.areas(area);
            let [area] = h.areas(area);

            frame.render_widget(Clear, area);
            frame.render_widget(table, area);
        }
        Ok(())
    }
}
