use crossterm::event::{KeyCode, KeyEvent, KeyEventKind};
use color_eyre::Result;
use ratatui::{
    layout::{Position, Alignment, Constraint, Flex, Layout, Rect}, 
    widgets::{Block, Clear, Paragraph}, 
    Frame
};
use tokio::sync::mpsc::UnboundedSender;

use super::{Component, PartialComponent, CurrentTab};
use crate::{
    widgets::{Input, InputRequest},
    action::Action, 
    config::Config, 
    palette::StylePalette, 
};

#[derive(Debug)]
enum NameOrAddress {
    Name,
    Address,
}

#[derive(Debug)]
pub struct AddAddressBookRecord {
    is_active: bool,
    name_or_address: NameOrAddress,
    action_tx: Option<UnboundedSender<Action>>,
    name: Input,
    address: Input,
    palette: StylePalette
}

impl Default for AddAddressBookRecord {
    fn default() -> Self {
        Self::new()
    }
}

impl AddAddressBookRecord {
    pub fn new() -> Self {
        Self {
            is_active: false,
            name_or_address: NameOrAddress::Name,
            action_tx: None,
            name: Input::new(String::new()),
            address: Input::new(String::new()),
            palette: StylePalette::default(),
        }
    }

    fn close_popup(&mut self) {
        self.is_active = false;
        if let Some(action_tx) = &self.action_tx {
            let _ = action_tx.send(Action::ClosePopup);
        }
    }

    fn submit_message(&mut self) {
        match self.name_or_address {
            NameOrAddress::Name => self.name_or_address = NameOrAddress::Address,
            NameOrAddress::Address => if let Some(action_tx) = &self.action_tx {
                let _ = action_tx.send(Action::NewAddressBookRecord(
                        self.name.value().to_string(),
                        self.address.value().to_string()));
                let _ = action_tx.send(Action::ClosePopup);
            }
        }
    }

    fn enter_char(&mut self, new_char: char) {
        match self.name_or_address {
            NameOrAddress::Name => {
                let _ = self.name.handle(InputRequest::InsertChar(new_char));
            },
            NameOrAddress::Address => {
                let _ = self.address.handle(InputRequest::InsertChar(new_char));
            }
        }
    }

    fn delete_char(&mut self) {
        match self.name_or_address {
            NameOrAddress::Name => {
                let _ = self.name.handle(InputRequest::DeletePrevChar);
            },
            NameOrAddress::Address => {
                let _ = self.address.handle(InputRequest::DeletePrevChar);
            }
        }
    }

    fn move_cursor_right(&mut self) {
        match self.name_or_address {
            NameOrAddress::Name => {
                let _ = self.name.handle(InputRequest::GoToNextChar);
            },
            NameOrAddress::Address => {
                let _ = self.address.handle(InputRequest::GoToNextChar);
            }
        }
    }

    fn move_cursor_left(&mut self) {
        match self.name_or_address {
            NameOrAddress::Name => {
                let _ = self.name.handle(InputRequest::GoToPrevChar);
            },
            NameOrAddress::Address => {
                let _ = self.address.handle(InputRequest::GoToPrevChar);
            }
        }
    }
}

impl PartialComponent for AddAddressBookRecord {
    fn set_active(&mut self, current_tab: CurrentTab) {
        match current_tab {
            CurrentTab::AddAddressBookRecord => self.is_active = true,
            _ => {
                if self.is_active {
                    self.is_active = false;
                    self.name = Input::new(String::new());
                    self.address = Input::new(String::new());
                    self.name_or_address = NameOrAddress::Name;
                }
            },
        };
    }
}

impl Component for AddAddressBookRecord {
    fn register_action_handler(&mut self, tx: UnboundedSender<Action>) -> Result<()> {
        self.action_tx = Some(tx);
        Ok(())
    }

    fn register_config_handler(&mut self, config: Config) -> Result<()> {
        if let Some(style) = config.styles.get(&crate::app::Mode::Wallet) {
            self.palette.with_normal_style(style.get("normal_style").copied());
            self.palette.with_normal_border_style(style.get("normal_border_style").copied());
            self.palette.with_normal_title_style(style.get("normal_title_style").copied());
            self.palette.with_popup_style(style.get("popup_style").copied());
            self.palette.with_popup_title_style(style.get("popup_title_style").copied());
        }
        Ok(())
    }

    fn handle_key_event(&mut self, key: KeyEvent) -> Result<Option<Action>> {
        if self.is_active && key.kind == KeyEventKind::Press {
            match key.code {
                KeyCode::Up => self.name_or_address = NameOrAddress::Name,
                KeyCode::Down => self.name_or_address = NameOrAddress::Address,
                KeyCode::Enter => self.submit_message(),
                KeyCode::Char(to_insert) => self.enter_char(to_insert),
                KeyCode::Backspace => self.delete_char(),
                KeyCode::Left => self.move_cursor_left(),
                KeyCode::Right => self.move_cursor_right(),
                KeyCode::Esc => self.close_popup(), 
                _ => {},
            };
        }
        Ok(None)
    }

    fn draw(&mut self, frame: &mut Frame, area: Rect) -> Result<()> {
        if self.is_active {
            let size = area.as_size();
            let name_area = Rect::new(size.width / 2, size.height / 2, 51, 3); 
            let address_area = Rect::new(size.width / 2, size.height / 2 + 3, 51, 3); 
            let (border_style, border_type) = self.palette.create_popup_style();

            let input_name = Paragraph::new(self.name.value())
                .block(Block::bordered()
                    .border_style(border_style)
                    .border_type(border_type)
                    .title_style(self.palette.create_popup_title_style())
                    .title_alignment(Alignment::Right)
                    .title("New name in Address Book"));

            let input_address = Paragraph::new(self.address.value())
                .block(Block::bordered()
                    .border_style(border_style)
                    .border_type(border_type)
                    .title_style(self.palette.create_popup_title_style())
                    .title_alignment(Alignment::Right)
                    .title("Address for new name"));

            let v = Layout::vertical([Constraint::Length(3)]).flex(Flex::Center);
            let h = Layout::horizontal([Constraint::Length(51)]).flex(Flex::Center);

            let [name_area] = v.areas(name_area);
            let [name_area] = h.areas(name_area);
            let [address_area] = v.areas(address_area);
            let [address_area] = h.areas(address_area);

            frame.render_widget(Clear, name_area);
            frame.render_widget(Clear, address_area);
            frame.render_widget(input_name, name_area);
            frame.render_widget(input_address, address_area);

            match self.name_or_address {
                NameOrAddress::Name => {
                    frame.set_cursor_position(Position::new(
                            name_area.x + self.name.cursor() as u16 + 1,
                            name_area.y + 1
                    ));
                },
                NameOrAddress::Address => {
                    frame.set_cursor_position(Position::new(
                            address_area.x + self.address.cursor() as u16 + 1,
                            address_area.y + 1
                    ));
                }
            }
        }
        Ok(())
    }
}
